@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: smtp.c,v 1.155 2003/08/20 05:12:58 motonori Exp $";
+static char *_id_ = "$Id: smtp.c,v 1.156 2003/08/21 07:04:31 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -76,7 +76,7 @@ sig_term()
 {
 	log(LOG_INFO, "SIGTERM received: semi graceful termination");
 	markstatus(env.rcpt_list, SMTP_TEMPFAIL(51),
-		"Aborted by INT signal", 1, 0);
+		"Aborted by INT signal", 1);
 	longjmp(jmp_env, 1);
 }
 
@@ -351,7 +351,7 @@ smtp()
 			/* DNS lookup failure or something else */
 			if (msg == NULL)
 				msg = "may be DNS problem";
-			markstatus(rcptp, SMTP_TEMPFAIL(52), msg, 0, 0);
+			markstatus(rcptp, SMTP_TEMPFAIL(52), msg, 0);
 			continue;
 		}
 		if (rcptp->domain == NULL)
@@ -528,7 +528,7 @@ smtp()
 			if (bad_state++ > 5)
 			{
 				markstatus(env.rcpt_list, SMTP_TEMPFAIL(51),
-					"No active sockets", 1, 0);
+					"No active sockets", 1);
 				dump_internal();
 				return;
 			}
@@ -663,7 +663,7 @@ new_transaction()
 						rc = SMTP_ERROR(52);
 					else
 						rc = SMTP_TEMPFAIL(52);
-					markstatus(rcptp, rc, msg, 0, 0);
+					markstatus(rcptp, rc, msg, 0);
 					rcptp->stat = RCPT_RETRY;
 				}
 				rcptp = rcptp->dom_chain;
@@ -740,7 +740,7 @@ new_transaction()
 #endif
 				rcode = SMTP_ERROR(50);
 		}
-		markstatus(conn->rcpt, rcode, msgbuf, 1, 0);
+		markstatus(conn->rcpt, rcode, msgbuf, 1);
 
 		close_transaction(conn, 1);
 		return 0;
@@ -869,12 +869,11 @@ int timeout_only, true_term;
 }
 
 static void
-markstatus(rcptp, rcode, response, tracechain, retryMX)
+markstatus(rcptp, rcode, response, tracechain)
 struct recipient *rcptp;
 int rcode;
 char *response;
 int tracechain;
-int retryMX;
 {
 	char hbuf[MAXLINE], rbuf[MAXLINE], pbuf[64];
 	char addrbuf[SYSLOG_BUFSIZE];
@@ -884,7 +883,11 @@ int retryMX;
 	static int PrevPCT = -1;
 
 	if (rcptp == NULL)
+	{
+		if (cnf.debug & DEBUG_TRANS)
+		log(LOG_DEBUG, "markstatus: rcptp is NULL");
 		return;
+	}
 
 	proto = "unknown";
 	delay = -1;
@@ -1021,8 +1024,8 @@ int retryMX;
 
 	while (rcptp != NULL)
 	{
-		if (rcptp->result != 0 && !IS4xx(rcptp->result)
-		 || rcptp->stat != RCPT_WORK)
+		if (IS2xx(rcptp->result) || IS5xx(rcptp->result)
+		 || IS4xx(rcptp->result) && rcptp->again && tracechain)
 		{
 			if (tracechain == 0)
 				break;
@@ -1062,7 +1065,7 @@ int retryMX;
 		{
 			struct mx *mxp = rcptp->domain->curmx;
 
-			if (retryMX)
+			if (rcptp->again)
 			{
 				rcptp->stat = RCPT_RETRY;
 				if (cnf.debug & DEBUG_TRANS)
@@ -1219,8 +1222,10 @@ struct recipient *rcptp;
 	connwork[i].crcptr = rcptp;
 	connwork[i].goodmail = 0;
 	connwork[i].goodrcpt = 0;
+#if 0
 	rcptp->stat = RCPT_WORK;
 	rcptp->chain = NULL;
+#endif
 	hostp = rcptp->domain->curmx->host;
 	connwork[i].host = hostp;
 	connwork[i].msg = env.msg;
@@ -1232,6 +1237,7 @@ struct recipient *rcptp;
 	rcpts = 1;
 	rcptp->stat = RCPT_WORK;
 	rcptp->chain = NULL;
+	rcptp->again = 0;
 	rcpt_tail = rcptp;
 	mxp = rcptp->domain->curmx;
 	first_time = 1;
@@ -1280,6 +1286,7 @@ struct recipient *rcptp;
 					sti.ntrans, rcptp2->address);
 				rcptp2->stat = RCPT_WORK;
 				rcptp2->chain = NULL;
+				rcptp2->again = 0;
 				rcpt_tail->chain = rcptp2;
 				rcpt_tail = rcptp2;
 			}
@@ -1338,9 +1345,11 @@ int next;	/* true: forward MX for next trial */
 
 	for (rcptp = conn->rcpt; rcptp != NULL; rcptp = rcptp->chain)
 	{
-		if (!Terminate && (rcptp->stat != RCPT_WORK))
+		if (!Terminate && (rcptp->stat == RCPT_DONE))
 			continue;
 		rcptp->stat = RCPT_RETRY;
+		if (rcptp->again)
+			continue;
 		domp = rcptp->domain;
 		if (domp->checked)
 			continue;
@@ -1633,7 +1642,7 @@ struct connection *conn;
 		conn->host->stat = SMTP_TEMPFAIL(52);
 		conn->host->response = "Connection was not established";
 		markstatus(conn->rcpt, conn->host->stat,
-			conn->host->response, 1, 0);
+			conn->host->response, 1);
 		return -1;
 	}
 
@@ -1837,7 +1846,7 @@ struct connection *conn;
 			conn->host->state = STAT_FAIL;
 			close (sd);
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(52),
-				"source port binding failure", 1, 0);
+				"source port binding failure", 1);
 			return -1;
 		}
 	}
@@ -1853,7 +1862,7 @@ struct connection *conn;
 		conn->host->state = STAT_FAIL;
 		close (sd);
 		markstatus(conn->rcpt, SMTP_TEMPFAIL(52),
-			"FIONBIO failed", 1, 0);
+			"FIONBIO failed", 1);
 		return -1;
 	}
 #else
@@ -1864,7 +1873,7 @@ struct connection *conn;
 		conn->host->state = STAT_FAIL;
 		close (sd);
 		markstatus(conn->rcpt, SMTP_TEMPFAIL(52),
-			"F_GETFL failed", 1, 0);
+			"F_GETFL failed", 1);
 		return -1;
 	}
 # ifdef O_NONBLOCK
@@ -1879,7 +1888,7 @@ struct connection *conn;
 		conn->host->state = STAT_FAIL;
 		close (sd);
 		markstatus(conn->rcpt, SMTP_TEMPFAIL(52),
-			"F_SETFL failed", 1, 0);
+			"F_SETFL failed", 1);
 		return -1;
 	}
 #endif
@@ -1895,7 +1904,7 @@ struct connection *conn;
 		conn->host->current->stat = EX_TEMPFAIL;
 		close (sd);
 		markstatus(conn->rcpt, SMTP_TEMPFAIL(52),
-			"connect() failed", 1, 0);
+			"connect() failed", 1);
 		/* try next address of the same MX */
 		conn->host->current = conn->host->current->next;
 #if 0
@@ -1977,7 +1986,7 @@ struct connection *conn;
 		  default:
 			errmsg = smtpstrerror(errno);
 		}
-		markstatus(conn->rcpt, SMTP_TEMPFAIL(52), errmsg, 1, 0);
+		markstatus(conn->rcpt, SMTP_TEMPFAIL(52), errmsg, 1);
 		if (conn->host->current != NULL) /* XXX for safe access */
 			conn->host->current = conn->host->current->next;
 		if (Terminate)
@@ -2236,7 +2245,7 @@ struct connection *conn;
 		conn->host->state = STAT_FAIL;
 		conn->host->stat = SMTP_TEMPFAIL(52);
 		markstatus(conn->rcpt, conn->host->stat,
-			conn->host->response, 1, 0);
+			conn->host->response, 1);
 		return -1;
 	}
 	return 0;
@@ -2279,7 +2288,7 @@ int *rest;
 			conn->host->state = STAT_FAIL;
 			conn->host->stat = SMTP_TEMPFAIL(52);
 			markstatus(conn->rcpt, conn->host->stat,
-				conn->host->response, 1, 0);
+				conn->host->response, 1);
 			return -1;
 		}
 		if (rcode == 0)
@@ -2332,7 +2341,7 @@ int *rest;
 		conn->host->response = newstr(msgbuf);
 		conn->host->state = STAT_FAIL;
 		markstatus(conn->rcpt, conn->host->stat,
-			conn->host->response, 1, 0);
+			conn->host->response, 1);
 		return -1;
 	}
 	return stat;
@@ -2367,7 +2376,7 @@ int flush;
 			conn->host->response
 				= "out of memory for sending buffer";
 			markstatus(conn->rcpt, conn->host->stat,
-				conn->host->response, 1, 0);
+				conn->host->response, 1);
 			return -1;
 		}
 		strcpy(p, buf);
@@ -2389,7 +2398,7 @@ int flush;
 		conn->host->state = STAT_FAIL;
 		conn->host->stat = SMTP_TEMPFAIL(52);
 		markstatus(conn->rcpt, conn->host->stat,
-			conn->host->response, 1, 0);
+			conn->host->response, 1);
 		return -1;
 	}
 	return rcode;
@@ -2473,7 +2482,7 @@ struct connection *conn;
 		conn->host->state = STAT_CLOSED;
 		if (errmsg != NULL)
 		{
-			markstatus(conn->rcpt, SMTP_TEMPFAIL(52), errmsg, 1, 0);
+			markstatus(conn->rcpt, SMTP_TEMPFAIL(52), errmsg, 1);
 			if (conn->host->current != NULL /* XXX for safe access */
 			 && conn->host->current->next == NULL
 			 && conn->host->response == NULL)
@@ -2492,7 +2501,7 @@ struct connection *conn;
 		{
 			sti.ntimeout++;
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-				"Processing timed out (greeting)", 1, 0);
+				"Processing timed out (greeting)", 1);
 			smtp_quit_s(conn);
 			return -1;
 		}
@@ -2659,7 +2668,7 @@ struct connection *conn;
 				conn->host->stat = SMTP_ERROR(53);
 				conn->host->response = "Config error: mail loops back to me (MX problem?)";
 				markstatus(conn->rcpt, conn->host->stat,
-					conn->host->response, 1, 0);
+					conn->host->response, 1);
 				smtp_quit_s(conn);
 				return -1;
 			}
@@ -2668,7 +2677,7 @@ struct connection *conn;
 		{
 			sti.ntimeout++;
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-				"Processing timed out (HELO)", 1, 0);
+				"Processing timed out (HELO)", 1);
 			smtp_quit_s(conn);
 			return -1;
 		}
@@ -2679,7 +2688,7 @@ struct connection *conn;
 			conn->host->stat = SMTP_TEMPFAIL(51);
 			conn->host->response = "Forced temporary failure";
 			markstatus(conn->rcpt, conn->host->stat,
-				 conn->host->response, 1, 0);
+				 conn->host->response, 1);
 			if (conn->reuse)
 				return -1;	/* without QUIT */
 			if (cnf.smtp_reuse && may_be_reused(conn))
@@ -2845,7 +2854,7 @@ struct connection *conn;
 	conn->host->state = STAT_FAIL;
 	conn->host->stat = SMTP_ERROR(54);
 	conn->host->response = "Sender address too long";
-	markstatus(conn->rcpt, conn->host->stat, conn->host->response, 1, 0);
+	markstatus(conn->rcpt, conn->host->stat, conn->host->response, 1);
 	smtp_quit_s(conn);
 	return -1;
 }
@@ -2887,13 +2896,13 @@ struct connection *conn;
 
 	if (IS2xx(rcode))
 	{
-		if (conn->host->proto & PROTO_PIPELINING)
-			conn->goodmail = 1;
+		/* if (conn->host->proto & PROTO_PIPELINING) */
+		conn->goodmail = 1;
 		if (Terminate)
 		{
 			sti.ntimeout++;
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-				"Processing timed out (MAIL)", 1, 0);
+				"Processing timed out (MAIL)", 1);
 			smtp_quit_s(conn);
 			return -1;
 		}
@@ -2918,7 +2927,7 @@ struct connection *conn;
 		else
 			rstr = newstr(buf + 4);
 	}
-	markstatus(conn->rcpt, rcode, rstr, 1, 0);	/* XXX */
+	markstatus(conn->rcpt, rcode, rstr, 1);	/* XXX */
 	conn->host->response = rstr;
 	conn->host->stat = rcode;
 
@@ -3008,7 +3017,7 @@ struct connection *conn;
 	conn->host->state = STAT_FAIL;
 	conn->host->stat = SMTP_ERROR(54);
 	conn->host->response = "Recipient address too long";
-	markstatus(conn->rcpt, conn->host->stat, conn->host->response, 1, 0);
+	markstatus(conn->rcpt, conn->host->stat, conn->host->response, 1);
 	smtp_quit_s(conn);	/* XXX */
 	return -1;
 }
@@ -3077,10 +3086,12 @@ struct connection *conn;
 		{
 			/* too many recipients */
 
-			if (!conn->reuse)
-				conn->reuse = 1;
-			/* conn->crcptr->stat = RCPT_RETRY; */
-
+			if (conn->rcpt != conn->crcptr) /* to avoid inf. loop */
+			{
+				conn->crcptr->again = 1;
+				if (!conn->reuse)
+					conn->reuse = 1;
+			}
 			if ((conn->host->proto & PROTO_PIPELINING) == 0)
 			{
 				/* preserve rest of recipients for next
@@ -3090,10 +3101,10 @@ struct connection *conn;
 				tmprcptp = conn->crcptr->chain;
 				while (tmprcptp != NULL)
 				{
-					tmprcptp->stat = RCPT_RETRY;
+					tmprcptp->again = 1;
+					conn->crcptr = tmprcptp; /* skip rest */
 					tmprcptp = tmprcptp->chain;
 				}
-				conn->crcptr->chain = NULL;
 			}
 		}
 
@@ -3113,10 +3124,7 @@ struct connection *conn;
 				rstr = newstr(skipaddr(buf + 4,
 					conn->crcptr->address));
 		}
-		if (rcode != SMTP_TEMPFAIL(52) || conn->rcpt == conn->crcptr)
-			markstatus(conn->crcptr, rcode, rstr, 0, 0);
-		else
-			markstatus(conn->crcptr, rcode, rstr, 0, 1);
+		markstatus(conn->crcptr, rcode, rstr, 0);
 	}
 	conn->crcptr = conn->crcptr->chain;
 
@@ -3124,7 +3132,7 @@ struct connection *conn;
 	{
 		sti.ntimeout++;
 		markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-			"Processing timed out (RCPT)", 1, 0);
+			"Processing timed out (RCPT)", 1);
 		smtp_quit_s(conn);
 		return -1;
 	}
@@ -3241,7 +3249,7 @@ struct connection *conn;
 			sti.ntimeout++;
 			conn->host->state = STAT_FAIL;
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-				"Processing timed out (DATA)", 1, 0);
+				"Processing timed out (DATA)", 1);
 			/* XXX how can i terminate here ? */
 			smtp_close(conn->socket);
 			if (conn->host->current->next == NULL)
@@ -3254,7 +3262,8 @@ struct connection *conn;
 		return smtp_data_body(conn);
 	}
 
-	if (rcode == SMTP_ERROR(03) || rcode == SMTP_ERROR(54))
+	if (conn->goodrcpt == 0
+	 && (rcode == SMTP_ERROR(03) || rcode == SMTP_ERROR(54)))
 	{
 		/* there were no valid recipients but we may have more
 		   recipients to attempt */
@@ -3285,7 +3294,7 @@ struct connection *conn;
 
 	if ((conn->host->proto & PROTO_PIPELINING) == 0 || conn->goodmail == 1)
 	{
-		markstatus(conn->rcpt, rcode, rstr, 1, 0);	/* XXX */
+		markstatus(conn->rcpt, rcode, rstr, 1);	/* XXX */
 		conn->host->stat = rcode;
 		conn->host->response = rstr;
 	}
@@ -3343,7 +3352,7 @@ struct connection *conn;
 				conn->host->stat = SMTP_TEMPFAIL(51);
 				conn->host->response = "lost connection";
 				markstatus(conn->rcpt, conn->host->stat,
-					conn->host->response, 1, 0);
+					conn->host->response, 1);
 				/* smtp_quit_s(conn); */
 				return -1;
 			}
@@ -3480,7 +3489,7 @@ struct connection *conn;
 	if (rcode == 221)
 		rcode = 250;	/* for MTAs which reply bogus code */
 
-	markstatus(conn->rcpt, rcode, rstr, 1, 0);
+	markstatus(conn->rcpt, rcode, rstr, 1);
 
 	if (IS2xx(rcode))
 	{
@@ -3564,7 +3573,7 @@ struct connection *conn;
 		{
 			sti.ntimeout++;
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-				"Processing timed out (RSET)", 1, 0);
+				"Processing timed out (RSET)", 1);
 			smtp_quit_s(conn);
 			return -1;
 		}
@@ -3680,7 +3689,7 @@ struct connection *conn;
 		 || conn->host->current == NULL /* XXX for safe access */
 		 || conn->host->current->next == NULL)
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(52),
-				newstr(buf), 1, 0);
+				newstr(buf), 1);
 		else
 			log(LOG_INFO, "Trying next address of %s",
 			    conn->host->name);
@@ -3717,7 +3726,7 @@ struct connection *conn;
 			conn->host->stat = EX_TEMPFAIL;
 			if (conn->rcpt != NULL)
 				markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-					newstr(buf), 1, 0);
+					newstr(buf), 1);
 		}
 		smtp_close(conn->socket);
 		return -1;
